#!/usr/bin/python3

"""
Diaphora, a diffing plugin for IDA
Copyright (c) 2015-2021, Joxean Koret

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#-------------------------------------------------------------------------------
# Use only for heuristics generating 1.0 ratios, results without false positives
HEUR_TYPE_NO_FPS = 0

# Use it for most heuristics; it will assign 1.0 ratios to the best chooser, 
# values between 0.5 and <1.0 to the specific partial chooser and <0.5 results
# to the unreliable chooser, if specified.
HEUR_TYPE_RATIO = 1

# Similar as before, but partial results are only assigned for matches with a
# min specified ratio.
HEUR_TYPE_RATIO_MAX = 2

#-------------------------------------------------------------------------------
HEUR_FLAG_NONE        = 0
HEUR_FLAG_UNRELIABLE  = 1
HEUR_FLAG_SLOW        = 2

#-------------------------------------------------------------------------------
HEURISTICS = []
HEURISTICS.append({
  "name":"Same RVA and hash",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Same RVA and hash' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as float) md1, cast(df.md_index as float) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where (df.rva = f.rva
                 or df.segment_rva = f.segment_rva)
               and df.bytes_hash = f.bytes_hash
               and df.instructions = f.instructions
               and ((f.name = df.name and substr(f.name, 1, 4) != 'sub_')
                 or (substr(f.name, 1, 4) = 'sub_' or substr(df.name, 1, 4)))""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same order and hash",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Same order and hash' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as float) md1, cast(df.md_index as float) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where df.id = f.id
               and df.bytes_hash = f.bytes_hash
               and df.instructions = f.instructions
               and ((f.name = df.name and substr(f.name, 1, 4) != 'sub_')
                 or (substr(f.name, 1, 4) = 'sub_' or substr(df.name, 1, 4)))
               and ((f.nodes > 1 and df.nodes > 1
                 and f.instructions > 5 and df.instructions > 5)
                  or f.instructions > 10 and df.instructions > 10)""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Function Hash",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Function Hash' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as float) md1, cast(df.md_index as float) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where f.function_hash = df.function_hash 
               and ((f.nodes > 1 and df.nodes > 1
                 and f.instructions > 5 and df.instructions > 5)
                  or f.instructions > 10 and df.instructions > 10)""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Bytes hash and names",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":"""  select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                    'Bytes hash and names' description,
                    f.pseudocode pseudo1, df.pseudocode pseudo2,
                    f.assembly asm1, df.assembly asm2,
                    f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                    f.nodes bb1, df.nodes bb2,
                    cast(f.md_index as float) md1, cast(df.md_index as float) md2
                from $MAIN$.functions f,
                     $DIFF$.functions df
               where f.bytes_hash = df.bytes_hash
                 and f.names = df.names
                 and f.names != '[]'
                 and f.instructions > 5 and df.instructions > 5""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Bytes hash",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Bytes hash' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as float) md1, cast(df.md_index as float) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where f.bytes_hash = df.bytes_hash
               and f.instructions > 5 and df.instructions > 5""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Bytes sum",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Bytes sum' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as float) md1, cast(df.md_index as float) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where f.bytes_sum = df.bytes_sum
               and f.size = df.size
               and f.mnemonics = df.mnemonics
               and f.instructions > 5 and df.instructions > 5""",
  "flags":HEUR_FLAG_UNRELIABLE
})

HEURISTICS.append({
  "name":"Equal assembly or pseudo-code",
  "category":"Best",
  "ratio":HEUR_TYPE_NO_FPS,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Equal pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.pseudocode = df.pseudocode
        and df.pseudocode is not null
        and f.pseudocode_lines >= 5
        and f.name not like 'nullsub%'
        and df.name not like 'nullsub%'
        $POSTFIX$
      union
     select f.address ea, f.name name1, df.address ea2, df.name name2, 'Equal pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.assembly = df.assembly
        and df.assembly is not null
        and f.instructions >= 4 and df.instructions >= 4
        and f.name not like 'nullsub%'
        and df.name not like 'nullsub%'
        $POSTFIX$ """,
  "flags":HEUR_FLAG_NONE|HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Same address and mnemonics",
  "category":"Best",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Same address and mnemonics' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as double) md1, cast(df.md_index as double) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where df.address = f.address
               and df.mnemonics = f.mnemonics
               and df.instructions = f.instructions
               and df.instructions > 5
               and ((f.name = df.name and substr(f.name, 1, 4) != 'sub_')
                 or (substr(f.name, 1, 4) = 'sub_' or substr(df.name, 1, 4)))""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same cleaned up assembly",
  "category":"Best",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Same cleaned up assembly' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.clean_assembly = df.clean_assembly
         and f.nodes > 3 and df.nodes > 3
         and f.name not like 'nullsub%'
         and df.name not like 'nullsub%'""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same cleaned pseudo-code",
  "category":"Best",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Same cleaned pseudo-code' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.clean_pseudo = df.clean_pseudo
         and f.pseudocode_lines > 5 and df.pseudocode_lines > 5
         and f.name not like 'nullsub%'
         and df.name not like 'nullsub%' $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same address, nodes, edges and mnemonics",
  "category":"Best",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same address, nodes, edges and mnemonics' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.rva = df.rva
        and f.instructions = df.instructions
        and f.nodes = df.nodes
        and f.edges = df.edges
        and f.mnemonics = df.mnemonics
        and f.nodes > 1""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same RVA",
  "category":"Best",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
                   'Equal RVA' description,
                   f.pseudocode pseudo1, df.pseudocode pseudo2,
                   f.assembly asm1, df.assembly asm2,
                   f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                   f.nodes bb1, df.nodes bb2,
                   cast(f.md_index as double) md1, cast(df.md_index as double) md2
              from $MAIN$.functions f,
                   $DIFF$.functions df
             where df.rva = f.rva
               $POSTFIX$""",
  "min":0.7,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same constants",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2,
            'Same constants' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.constants = df.constants
        and f.constants_count = df.constants_count
        and f.constants_count > 1 $POSTFIX$""",
  "min":0.5,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same KOKA hash and constants",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Same KOKA hash and constants' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.constants mc,
            $DIFF$.constants dc,
            $MAIN$.functions  f,
            $DIFF$.functions df
      where mc.constant = dc.constant
        and  f.id = mc.func_id
        and df.id = dc.func_id
        and f.kgh_hash = df.kgh_hash
        and f.nodes > 3 $POSTFIX$ """,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same rare KOKA hash",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same rare KOKA hash' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df,
            (select kgh_hash
               from $DIFF$.functions
              where kgh_hash != 0
              group by kgh_hash
             having count(*) <= 2
              union 
             select kgh_hash
               from $MAIN$.functions
              where kgh_hash != 0
              group by kgh_hash
             having count(*) <= 2
            ) shared_hashes
      where f.kgh_hash = df.kgh_hash
        and df.kgh_hash = shared_hashes.kgh_hash
        and f.nodes > 5 $POSTFIX$ """,
  "min":0.45,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same rare MD Index",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same rare MD Index' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df,
            (select md_index
               from $DIFF$.functions
              where md_index != 0
              group by md_index
             having count(*) <= 2
              union 
             select md_index
               from $MAIN$.functions
              where md_index != 0
              group by md_index
             having count(*) <= 2
            ) shared_mds
      where f.md_index = df.md_index
        and df.md_index = shared_mds.md_index
        and f.nodes > 10 $POSTFIX$ """,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same address and rare constant",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Same address and rare constant' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.constants mc,
            $DIFF$.constants dc,
            $MAIN$.functions  f,
            $DIFF$.functions df
      where mc.constant = dc.constant
        and  f.id = mc.func_id
        and df.id = dc.func_id
        and df.address = f.address""",
  "min":0.5,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same rare constant",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Same rare constant' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.constants mc,
            $DIFF$.constants dc,
            $MAIN$.functions  f,
            $DIFF$.functions df
      where mc.constant = dc.constant
        and  f.id = mc.func_id
        and df.id = dc.func_id
        and f.nodes > 3 and df.nodes > 3
        and f.constants_count > 0""",
  "min":0.2,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same MD Index and constants",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Same MD Index and constants' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2,
             df.tarjan_topological_sort, df.strongly_connected_spp
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.md_index = df.md_index
         and f.md_index > 0
         and f.nodes > 3 and df.nodes > 3
         and ((f.constants = df.constants
         and f.constants_count > 0)) $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"All or most attributes",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2,
            'All attributes' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.nodes = df.nodes 
        and f.edges = df.edges
        and f.indegree = df.indegree
        and f.outdegree = df.outdegree
        and f.size = df.size
        and f.instructions = df.instructions
        and f.mnemonics = df.mnemonics
        and f.names = df.names
        and f.prototype2 = df.prototype2
        and f.cyclomatic_complexity = df.cyclomatic_complexity
        and f.primes_value = df.primes_value
        and f.bytes_hash = df.bytes_hash
        and f.pseudocode_hash1 = df.pseudocode_hash1
        and f.pseudocode_primes = df.pseudocode_primes
        and f.pseudocode_hash2 = df.pseudocode_hash2
        and f.pseudocode_hash3 = df.pseudocode_hash3
        and f.strongly_connected = df.strongly_connected
        and f.loops = df.loops
        and f.tarjan_topological_sort = df.tarjan_topological_sort
        and f.strongly_connected_spp = df.strongly_connected_spp $POSTFIX$
      union 
     select f.address ea, f.name name1, df.address ea2, df.name name2,
            'Most attributes' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
       where f.nodes = df.nodes 
         and f.edges = df.edges
         and f.indegree = df.indegree
         and f.outdegree = df.outdegree
         and f.size = df.size
         and f.instructions = df.instructions
         and f.mnemonics = df.mnemonics
         and f.names = df.names
         and f.prototype2 = df.prototype2
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.primes_value = df.primes_value
         and f.bytes_hash = df.bytes_hash
         and f.strongly_connected = df.strongly_connected
         and f.loops = df.loops
         and f.tarjan_topological_sort = df.tarjan_topological_sort
         and f.strongly_connected_spp = df.strongly_connected_spp 
         $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Switch structures",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Switch structures' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.switches = df.switches
        and df.switches != '[]'
        and f.nodes > 5 and df.nodes > 5
        $POSTFIX$""",
  "min": 0.5,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Same address, nodes, edges and primes (re-ordered instructions)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2,
            'Same address, nodes, edges and primes (re-ordered instructions)' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.rva = df.rva
        and f.instructions = df.instructions
        and f.nodes = df.nodes
        and f.edges = df.edges
        and f.primes_value = df.primes_value
        and f.nodes > 3 $POSTFIX$""",
  "min":0.5,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Import names hash",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Import names hash' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.names = df.names
         and f.names != '[]'
         and f.md_index = df.md_index
         and f.instructions = df.instructions
         and f.nodes > 5 and df.nodes > 5 $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Nodes, edges, complexity, mnemonics, names, prototype, in-degree and out-degree",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select f.address ea, f.name name1, df.address ea2, df.name name2,
                     'Nodes, edges, complexity, mnemonics, names, prototype, in-degree and out-degree' description,
                     f.pseudocode pseudo1, df.pseudocode pseudo2,
                     f.assembly asm1, df.assembly asm2,
                     f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                     f.nodes bb1, df.nodes bb2,
                     cast(f.md_index as float) md1, cast(df.md_index as float) md2
                from $MAIN$.functions f,
                     $DIFF$.functions df
               where f.nodes = df.nodes
                 and f.edges = df.edges
                 and f.mnemonics = df.mnemonics
                 and f.names = df.names
                 and f.cyclomatic_complexity = df.cyclomatic_complexity
                 and f.prototype2 = df.prototype2
                 and f.indegree = df.indegree
                 and f.outdegree = df.outdegree
                 and f.nodes > 3
                 and f.edges > 3
                 and f.names != '[]'
                 $POSTFIX$
               union
              select f.address ea, f.name name1, df.address ea2, df.name name2,
                     'Nodes, edges, complexity, mnemonics, names and prototype' description,
                     f.pseudocode pseudo1, df.pseudocode pseudo2,
                     f.assembly asm1, df.assembly asm2,
                     f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
                     f.nodes bb1, df.nodes bb2,
                     cast(f.md_index as float) md1, cast(df.md_index as float) md2
                from $MAIN$.functions f,
                     $DIFF$.functions df
               where f.nodes = df.nodes
                 and f.edges = df.edges
                 and f.mnemonics = df.mnemonics
                 and f.names = df.names
                 and f.names != '[]'
                 and f.cyclomatic_complexity = df.cyclomatic_complexity
                 and f.prototype2 = df.prototype2
                 $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Mnemonics and names",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select f.address ea, f.name name1, df.address ea2, df.name name2,
             'Mnemonics and names' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.mnemonics = df.mnemonics
         and f.instructions = df.instructions
         and f.names = df.names
         and f.names != '[]'
         and f.instructions > 5 and df.instructions > 5
         $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Pseudo-code fuzzy (normal)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Pseudo-code fuzzy (normal)' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_hash1 = f.pseudocode_hash1
        and f.instructions > 5 and df.instructions > 5 """,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Pseudo-code fuzzy (mixed)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Pseudo-code fuzzy (mixed)' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_hash3 = f.pseudocode_hash3
        and f.instructions > 5 and df.instructions > 5 """,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Pseudo-code fuzzy (reverse)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Pseudo-code fuzzy (reverse)' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_hash2 = f.pseudocode_hash2
        and f.instructions > 5 and df.instructions > 5 """,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Pseudo-code fuzzy hash",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Pseudo-code fuzzy hash' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_hash1 = f.pseudocode_hash1
        and df.pseudocode_hash2 = f.pseudocode_hash2
        and df.pseudocode_hash3 = f.pseudocode_hash3
        and f.instructions > 5
        and df.instructions > 5 """,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Similar pseudo-code and names",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Similar pseudo-code and names' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.pseudocode_lines = df.pseudocode_lines
        and f.names = df.names
        and df.names != '[]'
        and df.pseudocode_lines > 5
        and df.pseudocode is not null 
        and f.pseudocode is not null
        $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Pseudo-code fuzzy AST hash",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Pseudo-code fuzzy AST hash' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_primes = f.pseudocode_primes
        and f.pseudocode_lines > 3
        and length(f.pseudocode_primes) >= 35
        $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Partial pseudo-code fuzzy hash (normal)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""  select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Partial pseudo-code fuzzy hash (normal)' description,
              f.pseudocode pseudo1, df.pseudocode pseudo2,
              f.assembly asm1, df.assembly asm2,
              f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
              f.nodes bb1, df.nodes bb2,
              cast(f.md_index as float) md1, cast(df.md_index as float) md2
         from $MAIN$.functions f,
              $DIFF$.functions df
        where substr(df.pseudocode_hash1, 1, 16) = substr(f.pseudocode_hash1, 1, 16)
          and f.nodes > 5 and df.nodes > 5""",
  "min":0.5,
  "flags":HEUR_FLAG_SLOW
})


HEURISTICS.append({
  "name":"Partial pseudo-code fuzzy hash (reverse)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""  select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Partial pseudo-code fuzzy hash (reverse)' description,
              f.pseudocode pseudo1, df.pseudocode pseudo2,
              f.assembly asm1, df.assembly asm2,
              f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
              f.nodes bb1, df.nodes bb2,
              cast(f.md_index as float) md1, cast(df.md_index as float) md2
         from $MAIN$.functions f,
              $DIFF$.functions df
        where substr(df.pseudocode_hash2, 1, 16) = substr(f.pseudocode_hash2, 1, 16)
          and f.nodes > 5 and df.nodes > 5""",
  "min":0.5,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Partial pseudo-code fuzzy hash (mixed)",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""  select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Partial pseudo-code fuzzy hash (mixed)' description,
              f.pseudocode pseudo1, df.pseudocode pseudo2,
              f.assembly asm1, df.assembly asm2,
              f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
              f.nodes bb1, df.nodes bb2,
              cast(f.md_index as float) md1, cast(df.md_index as float) md2
         from $MAIN$.functions f,
              $DIFF$.functions df
        where substr(df.pseudocode_hash3, 1, 16) = substr(f.pseudocode_hash3, 1, 16)
          and f.nodes > 5 and df.nodes > 5""",
  "min":0.5,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Strongly connected components small-primes-product",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""  select f.address ea, f.name name1, df.address ea2, df.name name2, 'Strongly connected components small-primes-product' description,
              f.pseudocode pseudo1, df.pseudocode pseudo2,
              f.assembly asm1, df.assembly asm2,
              f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
              f.nodes bb1, df.nodes bb2,
              cast(f.md_index as float) md1, cast(df.md_index as float) md2
         from $MAIN$.functions f,
              $DIFF$.functions df
        where f.strongly_connected_spp = df.strongly_connected_spp
          and df.strongly_connected_spp > 1
          and f.nodes > 10 and df.nodes > 10
          $POSTFIX$""",
  "flags":HEUR_FLAG_NONE|HEUR_FLAG_UNRELIABLE
})

HEURISTICS.append({
  "name":"Loop count",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Loop count' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.loops = df.loops
        and df.loops > 1
        and f.nodes > 3 and df.nodes > 3
        $POSTFIX$""",
  "min":0.49,
  "flags":HEUR_FLAG_SLOW
})


HEURISTICS.append({
  "name":"Mnemonics small-primes-product",
  "category":"Partial",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":""" select f.address ea, f.name name1, df.address ea2, df.name name2,
             'Mnemonics small-primes-product' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.mnemonics_spp = df.mnemonics_spp
         and f.instructions = df.instructions
         and f.nodes > 1 and df.nodes > 1
         and df.instructions > 5 $POSTFIX$ """,
  "min":0.6,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Similar pseudo-code",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Similar pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.pseudocode_lines = df.pseudocode_lines
        and df.pseudocode_lines > 5
        and df.pseudocode is not null 
        and f.pseudocode is not null
        $POSTFIX$""",
  "min":0.6,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same nodes, edges and strongly connected components",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2,
            'Same nodes, edges and strongly connected components' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.nodes = df.nodes
        and f.edges = df.edges
        and f.strongly_connected = df.strongly_connected
        and f.nodes > 5 and df.nodes > 5
        $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Similar small pseudo-code",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Similar small pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.pseudocode_lines = df.pseudocode_lines
        and df.pseudocode_lines <= 5
        and df.pseudocode is not null 
        and f.pseudocode is not null
        $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Small pseudo-code fuzzy AST hash",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select distinct f.address ea, f.name name1, df.address ea2, df.name name2, 'Small pseudo-code fuzzy AST hash' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode_primes = f.pseudocode_primes
        and f.pseudocode_lines <= 5""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Equal small pseudo-code",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Equal small pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.pseudocode = df.pseudocode
        and df.pseudocode is not null
        and f.pseudocode_lines < 5
        $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same low complexity, prototype and names",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""  select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same low complexity, prototype and names' description,
              f.pseudocode pseudo1, df.pseudocode pseudo2,
              f.assembly asm1, df.assembly asm2,
              f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
              f.nodes bb1, df.nodes bb2,
              cast(f.md_index as float) md1, cast(df.md_index as float) md2
         from $MAIN$.functions f,
              $DIFF$.functions df
        where f.names = df.names
          and f.cyclomatic_complexity = df.cyclomatic_complexity
          and f.cyclomatic_complexity < 20
          and f.prototype2 = df.prototype2
          and df.names != '[]'
          $POSTFIX$""",
  "min":0.5,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same low complexity and names",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same low complexity and names' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.names = df.names
        and f.cyclomatic_complexity = df.cyclomatic_complexity
        and f.cyclomatic_complexity < 15
        and df.names != '[]'
        $POSTFIX$""",
  "min":0.5,
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Same graph",
  "category":"Experimental",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":""" select f.address ea, f.name name1, df.address ea2, df.name name2,
             'Same graph' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.nodes = df.nodes 
         and f.edges = df.edges
         and f.indegree = df.indegree
         and f.outdegree = df.outdegree
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.strongly_connected = df.strongly_connected
         and f.loops = df.loops
         and f.tarjan_topological_sort = df.tarjan_topological_sort
         and f.strongly_connected_spp = df.strongly_connected_spp
         $POSTFIX$
         and f.nodes > 5 and df.nodes > 5
       order by
             case when f.size = df.size then 1 else 0 end +
             case when f.instructions = df.instructions then 1 else 0 end +
             case when f.mnemonics = df.mnemonics then 1 else 0 end +
             case when f.names = df.names then 1 else 0 end +
             case when f.prototype2 = df.prototype2 then 1 else 0 end +
             case when f.primes_value = df.primes_value then 1 else 0 end +
             case when f.bytes_hash = df.bytes_hash then 1 else 0 end +
             case when f.pseudocode_hash1 = df.pseudocode_hash1 then 1 else 0 end +
             case when f.pseudocode_primes = df.pseudocode_primes then 1 else 0 end +
             case when f.pseudocode_hash2 = df.pseudocode_hash2 then 1 else 0 end +
             case when f.pseudocode_hash3 = df.pseudocode_hash3 then 1 else 0 end DESC""",
  "min":0.5,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Topological sort hash",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2,
            'Topological sort hash' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.strongly_connected = df.strongly_connected
        and f.tarjan_topological_sort = df.tarjan_topological_sort
        and f.strongly_connected > 3
        and f.nodes > 10
        $POSTFIX$""",
  "flags":HEUR_FLAG_NONE
})

HEURISTICS.append({
  "name":"Strongly connected components SPP and names",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":""" select f.address ea, f.name name1, df.address ea2, df.name name2,
             'Strongly connected components SPP and names' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.names = df.names
         and f.names != '[]'
         and f.strongly_connected_spp = df.strongly_connected_spp
         and f.strongly_connected_spp > 0
         and f.nodes > 5 and df.nodes > 5 """,
  "min":0.49,
  "flags":HEUR_FLAG_UNRELIABLE
})

HEURISTICS.append({
  "name":"Strongly connected components",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Strongly connected components' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.strongly_connected = df.strongly_connected
        and df.strongly_connected > 1
        and f.nodes > 5 and df.nodes > 5
        and f.strongly_connected_spp > 1
        and df.strongly_connected_spp > 1
        $POSTFIX$""",
  "min":0.8,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Loop count",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Loop count' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.loops = df.loops
        and df.loops > 1
        $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Nodes, edges, complexity and mnemonics",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Nodes, edges, complexity and mnemonics' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.nodes = df.nodes
         and f.edges = df.edges
         and f.mnemonics = df.mnemonics
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.nodes > 1 and f.edges > 0
         $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Nodes, edges, complexity and prototype",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Nodes, edges, complexity and prototype' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.nodes = df.nodes
         and f.edges = df.edges
         and f.prototype2 = df.prototype2
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.prototype2 != 'int()'
         $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Nodes, edges, complexity, in-degree and out-degree",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Nodes, edges, complexity, in-degree and out-degree' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.nodes = df.nodes
         and f.edges = df.edges
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.nodes > 3 and f.edges > 2
         and f.indegree = df.indegree
         and f.outdegree = df.outdegree
         $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Nodes, edges and complexity",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":""" select distinct f.address ea, f.name name1, df.address ea2, df.name name2,
             'Nodes, edges and complexity' description,
             f.pseudocode pseudo1, df.pseudocode pseudo2,
             f.assembly asm1, df.assembly asm2,
             f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
             f.nodes bb1, df.nodes bb2,
             cast(f.md_index as float) md1, cast(df.md_index as float) md2
        from $MAIN$.functions f,
             $DIFF$.functions df
       where f.nodes = df.nodes
         and f.edges = df.edges
         and f.cyclomatic_complexity = df.cyclomatic_complexity
         and f.nodes > 1 and f.edges > 0
         $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Similar small pseudo-code",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO_MAX,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Similar small pseudo-code' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where df.pseudocode is not null 
        and f.pseudocode is not null
        and f.pseudocode_lines = df.pseudocode_lines
        and df.pseudocode_lines > 5
        $POSTFIX$""",
  "min":0.5,
  "flags":HEUR_FLAG_SLOW
})

HEURISTICS.append({
  "name":"Same high complexity",
  "category":"Unreliable",
  "ratio":HEUR_TYPE_RATIO,
  "sql":"""select f.address ea, f.name name1, df.address ea2, df.name name2, 'Same high complexity' description,
            f.pseudocode pseudo1, df.pseudocode pseudo2,
            f.assembly asm1, df.assembly asm2,
            f.pseudocode_primes pseudo_primes1, df.pseudocode_primes pseudo_primes2,
            f.nodes bb1, df.nodes bb2,
            cast(f.md_index as float) md1, cast(df.md_index as float) md2
       from $MAIN$.functions f,
            $DIFF$.functions df
      where f.cyclomatic_complexity = df.cyclomatic_complexity
        and f.cyclomatic_complexity >= 50
        $POSTFIX$""",
  "flags":HEUR_FLAG_SLOW
})

#-------------------------------------------------------------------------------
def check_categories():
  categories = set()
  for heur in HEURISTICS:
    category = heur["category"]
    categories.add(category)

  print("Categories:")
  import pprint
  pprint.pprint(categories)
  
  assert(categories == set(['Best', 'Experimental', 'Partial', 'Unreliable']))

#-------------------------------------------------------------------------------
def check_dupes():
  from collections import Counter
  heurs = Counter()
  for heur in HEURISTICS:
    if "name" not in heur:
      print("No 'name' field in heuristic!")
      print(heur)
      assert(name in heur)

    name = heur["name"]
    heurs[name] += 1

  dups = []
  for key in heurs:
    if heurs[key] > 1:
      dups.append([key, heurs[key]])
  
  print("Dups:")
  import pprint
  pprint.pprint(dups)
  
  assert(sorted(dups) == sorted([['Similar small pseudo-code', 2], ['Loop count', 2]]) )
  
#-------------------------------------------------------------------------------
def check_heuristic_in_sql():
  heurs = set()
  excluded = ["Equal assembly or pseudo-code", "All or most attributes"]
  for heur in HEURISTICS:
    name = heur["name"]
    if name in excluded:
      continue

    sql = heur["sql"]
    if sql.lower().find(name.lower()) == -1:
      print(("SQL command not correctly associated to %s" % repr(name)))
      print(sql)
      assert(sql.find(name) != -1)

    heurs.add(name)

  print("Heuristics:")
  import pprint
  pprint.pprint(heurs)

#-------------------------------------------------------------------------------
def check_heuristics_ratio():
  from collections import Counter
  ratios = Counter()
  for heur in HEURISTICS:
    if "ratio" not in heur:
      print("No 'ratio' in heuristic!")
      print(heur)
      assert("ratio" in heur)

    ratio = heur["ratio"]
    ratios[ratio] += 1
  
  print("Ratios:")
  import pprint
  pprint.pprint(ratios)
  
  assert(ratios == Counter({1: 29, 2: 18, 0: 7}))

#-------------------------------------------------------------------------------
def check_mandatory_fields():
  mandatory = set(["name", "ratio", "category", "sql", "flags"])
  for heur in HEURISTICS:
    for field in mandatory:
      if field not in list(heur.keys()):
        print(("Field '%s' not found in heuristic!" % field))
        print(heur)
        assert(field in list(heur.keys()))

#-------------------------------------------------------------------------------
def check_field_names():
  expected = set(["name", "ratio", "category", "min", "sql", "flags"])
  fields = set()
  for heur in HEURISTICS:
    for field in list(heur.keys()):
      if field not in expected:
        print(("Invalid field '%s' found for heuristic!" % field))
        print(heur)
        assert(field in expected)
      
      if heur["ratio"] == HEUR_TYPE_RATIO_MAX:
        if "min" not in heur:
          print("Heuristic of type HEUR_TYPE_RATIO_MAX without a minimum value set!")
          print(heur)
          assert("min" in heur)

      fields.add(field)

  import pprint
  pprint.pprint(fields)

#-------------------------------------------------------------------------------
def run_tests():
  print("Running tests...\n")
  check_categories()
  print("")
  check_dupes()
  print("")
  check_heuristic_in_sql()
  print("")
  check_heuristics_ratio()
  print("")
  check_field_names()
  print("")
  check_mandatory_fields()
  print("\nAll tests run OK!")

if __name__ == "__main__":
  run_tests()




__author__ = "Fernando Domínguez"
__credits__ = "Joxean Koret"

from .diaphora_r2 import generate_db_for_file, dbname_for_file, compare_dbs
from .diaphora import (
    get_function_details, get_db_attrs, get_db_attrs_path, drop_all,
    drop_db, db_exists as sample_db_exists
)
from .idaapi.idaapi_to_r2 import get_all_fns, strings, string_values, r2_open, r2_close
from .html_diff import HtmlResults





import magic
import difflib
from yattag import Doc
from hashlib import sha256

from r2diaphora import get_function_details
from .difflibparser import DifflibParser, DiffCode

class HtmlResults():

    results = []
    MATCHES_COLUMNS = ["type", "name", "address", "bb1", "name2", "address2", "bb2", "ratio", "description"]
    MATCHES_COLUMN_NAMES = {
        "address": "Address",
        "name2": "Name 2",
        "address2": "Address 2",
        "bb1": "BB1",
        "bb2": "BB2"
    }

    def __init__(self, results, file1 = None, file2 = None):
        self.results = results
        self.file1 = file1
        self.file2 = file2
        self.hashes = {}

    def get_file_hash(self, file):
        if file in self.hashes:
            return self.hashes[file]

        hash = ""
        with open(file, "rb") as f:
            d = f.read()
            hash = sha256(d).hexdigest();
        # Cache the result to not compute it multiple times
        self.hashes[file] = hash
        return hash

    def get_file_magic(self, file):
        m = magic.Magic()
        return m.from_file(file)

    def interpolate_color(self, color_a, color_b, alpha):
        # color_a and color_b should be tuples of 3 ints
        r_r = (color_b[0] - color_a[0]) * alpha + color_a[0]
        r_g = (color_b[1] - color_a[1]) * alpha + color_a[1]
        r_b = (color_b[2] - color_a[2]) * alpha + color_a[2]
        return (r_r, r_g, r_b)

    def render(self, filepath):
        doc, tag, text = Doc().tagtext()
        with tag("style"):
            doc.asis("""
            .code { 
                font-family: monospace;
                background-color: #ddd;
                padding: 20px;
                border-radius: 5px;
                color: #666;
            }

            table.table {
                margin: 15px;
            }

            table.table-dark {
                --bs-table-bg: #303538;
                --bs-table-striped-bg: #303538;
            }

            table.table tbody {
                font-family: monospace;
                font-size: 16px;
            }

            table.table>tbody>tr {
                color: #ccc;
                line-height: 30px;
            }

            .green {
                color: rgb(126, 211, 33);
            }

            .orange {
                color: rgb(245, 166, 35);
            }

            .yellow {
                color: rgb(216, 203, 42);
            }

            .red {
                color: rgb(255, 65, 89);
            }

            i.bi {
                color: #f44336;
            }

            .hidden-row {
                padding: 0 !important;
            }

            .pseudocode-tab {
                position: relative;
                border-radius: 20px;
                margin-top: 30px;
            }

            .pseudocode-tab > span {
                display: block;
                width: 100%;
                overflow: hidden;
            }

            span.equal {
                background-color: #0d1117;
            }

            span.blank {
                background-color: #0d1117ab;
            }

            span.deleted {
                background-color: #301a1f;
            }

            span.added {
                background-color: #12261e;
            }

            .nav-pills a.nav-link.active, .nav-pills .show>a.nav-link {
                background-color: #e91e63;
            }

            a.nav-link {
                color: #e91e63;
            }

            a.nav-link:focus, a.nav-link:hover {
                color: #ee4f84;
            }

            .table-hover>tbody>tr.nohover:hover>* {
                color: inherit;
                --bs-table-accent-bg: inherit;
            }

            i.bi {
                cursor: pointer;
            }
            """)

        with tag("html"):
            with tag("head"):
                doc.stag("link", rel="stylesheet", href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.2/dist/css/bootstrap.min.css")
                doc.stag("link", rel="stylesheet", href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.5.0/font/bootstrap-icons.css")
                with tag("script", src="https://code.jquery.com/jquery-3.6.0.slim.min.js"):
                    pass
                with tag("script", src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.2/dist/js/bootstrap.bundle.min.js"):
                    pass

            with tag("body", style="margin: 50px; background-color: #303538; color: white;"):
                for i, f in enumerate([self.file1, self.file2]):
                    if not f:
                        continue

                    with tag("div", klass="row"):
                        with tag("div", klass="col-md-12"):
                            with tag("p", klass="code"):
                                text(f"File {i + 1}: {f}")
                                doc.stag("br")
                                text(f"Type: {self.get_file_magic(f)}")
                                doc.stag("br")
                                text(f"SHA256: {self.get_file_hash(f)}")

                with tag("h5"):
                    text(f"Found {len(self.results)} matches across compared files")

                with tag("div", klass="row"):
                    with tag("table", 
                            klass="table table-dark table-hover align-middle",
                            style="width: 98%;"):
                        with tag("thead", klass="thead-dark"):
                            with tag("tr"):
                                doc.stag("th")
                                for attr in self.MATCHES_COLUMNS:
                                    with tag("th"):
                                        text(self.MATCHES_COLUMN_NAMES.get(attr, attr.capitalize()))
                        with tag("tbody"):
                            for i, r in enumerate(self.results):
                                # Match row
                                with tag("tr"):
                                    # Expand button
                                    with tag("th"):
                                        with tag("i",
                                                ("data-bs-toggle", "collapse"),
                                                ("data-bs-taget", f"#diff-{i}"),
                                                klass="collapse-control",
                                                href=f"#diff-{i}"):
                                            doc.stag("i", klass="bi bi-chevron-bar-expand")
                                    # Info columns
                                    for attr in self.MATCHES_COLUMNS:
                                        style = ""
                                        if attr == "ratio":
                                            ratio = float(r[attr])
                                            color = self.interpolate_color((255, 65, 89), (126, 211, 33), ratio)
                                            style = f"color: rgb({color[0]}, {color[1]}, {color[2]})"

                                        with tag("th", style=style):
                                                text(f"0x{r[attr]}" if attr in ["address", "address2"] else r[attr])
                                    # /Info columns
                                # Diff
                                with tag("tr", klass="nohover"):
                                    with tag("td", colspan=12, klass="hidden-row"):
                                        with tag("div", klass="accordion-body collapse", id=f"diff-{i}"):
                                            details1 = get_function_details(self.get_file_hash(self.file1), r["name"])
                                            pseudo1 = None
                                            if details1["prototype"] and details1["pseudocode"]:
                                                pseudo1 = details1["prototype"] + "\n" + details1["pseudocode"]
                                            details2 = get_function_details(self.get_file_hash(self.file2), r["name2"])
                                            pseudo2 = None
                                            if details2["prototype"] and details2["pseudocode"]:
                                                pseudo2 = details2["prototype"] + "\n" + details2["pseudocode"]

                                            changes = []
                                            if pseudo1 and pseudo2:
                                                diff = DifflibParser(
                                                    [f"{l}\n" for l in pseudo1.split("\n")],
                                                    [f"{l}\n" for l in pseudo2.split("\n")],
                                                )
                                                changes = [d for d in diff]

                                            # Tabs
                                            with tag("nav", klass="nav nav-pills justify-content-center", style="margin-bottom: 20px;"):
                                                for j, tab in enumerate(["Pseudocode", "Assembly", "Callgraph"]):
                                                    cls = "nav-link"
                                                    if j == 0:
                                                        cls += " active"
                                                    with tag("a", ("data-toggle", "tab"), klass=cls, href=f"#diff-{i}-{tab.lower()}"):
                                                        text(tab)
                                            # Tabs content
                                            with tag("div", klass="tab-content"):
                                                for j, tab in enumerate(["Pseudocode", "Assembly", "Callgraph"]):
                                                    cls = "tab-pane"
                                                    if j == 0:
                                                        cls += " active"

                                                    if tab == "Pseudocode":
                                                        if not pseudo1 or not pseudo2:
                                                            cls += " disabled"

                                                        with tag("div", klass=cls, id=f"diff-{i}-{tab.lower()}"):
                                                            with tag("div", klass="row"):
                                                                for side in [DiffCode.LEFTONLY, DiffCode.RIGHTONLY]:
                                                                    cls = "col-md-6"
                                                                    with tag("pre", klass=f"col-md-6 {tab.lower()}-tab"):
                                                                        for change in changes:
                                                                            if change["code"] == DiffCode.SIMILAR:
                                                                                with tag("span", klass="equal"):
                                                                                    text(change["line"])
                                                                            elif change["code"] == side:
                                                                                with tag(
                                                                                    "span",
                                                                                    klass=f"{'deleted' if side == DiffCode.LEFTONLY else 'added'}"
                                                                                ):
                                                                                    text(change["line"])
                                                                            elif change["code"] == DiffCode.CHANGED and side == DiffCode.LEFTONLY:
                                                                                with tag("span", klass="deleted"):
                                                                                    text(change["line"])
                                                                            elif change["code"] == DiffCode.CHANGED and side == DiffCode.RIGHTONLY:
                                                                                with tag("span", klass="added"):
                                                                                    text(change["newline"])
                                                                            else:
                                                                                with tag("span", klass="blank"):
                                                                                    text("\n")

                                                    elif tab == "Assembly":
                                                        pass

                                                    elif tab == "Callgraph":
                                                        pass


        result = doc.getvalue()
        with open(filepath, "w") as f:
            f.write(result)



#!/usr/bin/python

"""
Yet another Control Flow Graph hash using small-primes-product.
An implementation of the Koret-Karamitas (KOKA) CFGs hashing algorithm.

Based on the paper Efficient Features for Function Matching between Binary
Executables by Huku (Chariton Karamitas, CENSUS S.A., huku@census-labs.com).

Copyright (c) 2018-2019, Joxean Koret
"""



import sys
import time
import logging

from r2diaphora.idaapi.idaapi_to_r2 import *

try:
  from r2diaphora.others.tarjan_sort import strongly_connected_components
except ImportError:
  from r2diaphora.tarjan_sort import strongly_connected_components

log = logging.getLogger("diaphora.graph_hashes")

#-------------------------------------------------------------------------------
# Different type of basic blocks (graph nodes).
NODE_ENTRY = 2
NODE_EXIT = 3
NODE_NORMAL = 5

#
# NOTE: In the current implementation (Nov-2018) all edges are considered as if
# they were conditional. Keep reading...
#
EDGE_IN_CONDITIONAL = 7
EDGE_OUT_CONDITIONAL = 11

#
# Reserved but unused because, probably, it doesn't make sense when comparing
# multiple different architectures.
#
#EDGE_IN_UNCONDITIONAL = 13
#EDGE_OUT_UNCONDITIONAL = 17

# 
# The following are feature types that aren't applied at basic block but rather
# at function level. The idea is that if we do at function level we will have no
# problems finding the same function that was re-ordered because of some crazy
# code a different compiler decided to create (i.e., resilient to reordering).
#
FEATURE_LOOP = 19
FEATURE_CALL = 23
FEATURE_DATA_REFS = 29
FEATURE_CALL_REF = 31
FEATURE_STRONGLY_CONNECTED = 37
FEATURE_FUNC_NO_RET = 41
FEATURE_FUNC_LIB = 43
FEATURE_FUNC_THUNK = 47

#-------------------------------------------------------------------------------
# Implementation of the KOKA (Koret-Karamitas) hashing algorithm for IDA
class CKoretKaramitasHash:
  def __init__(self, r2):
    self.r2 = r2

  #-------------------------------------------------------------------------------
  def log_exec_r2_cmdj(self, cmd):
    s = time.time()
    r = self.r2.cmdj(cmd)
    log.debug(f"R2 CMDJ: {cmd}: {time.time() - s}s")
    return r

  def log_exec_r2_cmd(self, cmd):
    s = time.time()
    r = self.r2.cmd(cmd)
    log.debug(f"R2 CMD: {cmd}: {time.time() - s}s")
    return r

  #-------------------------------------------------------------------------------
  def get_node_value(self, succs, preds):
    """ Return a set of prime numbers corresponding to the characteristics of the node. """
    ret = 1
    if preds == 0:
      ret *= NODE_ENTRY
    
    if succs == 0:
      ret *= NODE_EXIT

    ret *= NODE_NORMAL
    return ret

  def get_edges_value(self, bb, succs, preds):
    ret = 1
    for _ in succs:
      ret *= EDGE_OUT_CONDITIONAL

    for _ in preds:
      ret *= EDGE_IN_CONDITIONAL

    return ret

  def is_call_insn(self, ea):
    _, inss = diaphora_decode(ea)
    if len(inss) < 1:
      return False
    return "call" in inss[0].get("type")

  def calculate(self, f):
    fname = get_function_name(f).get("name")
    if not fname:
      return "NO-FUNCTION"

    flow = self.log_exec_r2_cmdj(f"afbj @ {f}")
    if flow is None:
      return "NO-FLOW-GRAPH"

    hash = 1

    # Variables required for calculations of previous ones
    bb_relations = {}

    # Iterate through each basic block
    for block in flow:
      block_ea = +block['addr'];
      block_end_ea = +block['addr'] + +block['size'];
      block.update({"start_ea": block_ea})
      block.update({"end_ea": block_end_ea})

      if block["end_ea"] == 0:
        continue

      succs = block_succs(block_ea)
      preds = block_preds(block_ea)

      hash *= self.get_node_value(len(succs), len(preds))
      hash *= self.get_edges_value(block, succs, preds)

      # ...and each instruction on each basic block
      for ea in Heads(block["start_ea"], block["ninstr"]):

        if self.is_call_insn(ea):
          hash *= FEATURE_CALL

        l = list(DataRefsFrom(ea))
        if len(l) > 0:
          hash *= FEATURE_DATA_REFS

        for xref in CodeRefsFrom(ea, 0):
          if not is_func(xref) or get_flag_at_addr(xref).get("name") != fname:
            hash *= FEATURE_CALL_REF

        # Remember the relationships
        bb_relations[block_ea] = []

        # Iterate the succesors of this basic block
        for succ_block_ea in succs:
          bb_relations[block_ea].append(succ_block_ea)

        # Iterate the predecessors of this basic block
        for pred_block_ea in preds:
          try:
            bb_relations[pred_block_ea].append(block["start_ea"])
          except KeyError:
            bb_relations[pred_block_ea] = [block["start_ea"]]

    # Calculate the strongly connected components
    try:
      strongly_connected = strongly_connected_components(bb_relations)
      # ...and get the number of loops out of it
      for sc in strongly_connected:
        if len(sc) > 1:
          hash *= FEATURE_LOOP
        else:
          if sc[0] in bb_relations and sc[0] in bb_relations[sc[0]]:
            hash *= FEATURE_LOOP

      # And, also, use the number of strongly connected components
      # to calculate another part of the hash.
      hash *= (FEATURE_STRONGLY_CONNECTED ** len(strongly_connected))
    except:
      print("Exception:", str(sys.exc_info()[1]))

    if fname in no_ret_functions():
      hash *= FEATURE_FUNC_NO_RET
    if fname.startswith("flirt."):
      hash *= FEATURE_FUNC_LIB
    # if flags & FUNC_THUNK:
    #   hash *= FEATURE_FUNC_THUNK

    return str(hash)

#-------------------------------------------------------------------------------
def main():
  kgh = CKoretKaramitasHash()

  d = {}
  for f in Functions():
    hash = kgh.calculate(f)
    func_str_ea = "0x%08x" % f
    try:
      d[hash].append(func_str_ea)
    except:
      d[hash] = [func_str_ea]

    print("0x%08x %s" % (f, hash))

  import pprint
  pprint.pprint(d)
  uniques = 0
  for key in d:
    if len(d[key]) > 1:
      print(key, d[key])
    else:
      uniques += 1
  
  print()
  print("Unique hashes", uniques)

if __name__ == "__main__":
  main()
